unit MT_MainForm;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, Buttons,
  StdCtrls, LCLProc, LCLType, LCLIntf, Grids,
  mt_iter, mt_Logger;

type
  TResult = record
    value: double;
    root: double;
    iterations: integer;
  end;

  //TRoot = class
  //  xvalue: double;
  //  prec: double;
  //  root: double;
  //  iterations: integer;
  //end;

type
  { TMyThread }
  TMyThread = class(TThread)
  private
    FMyNumber: double;
    FMyResult: double;
    FMyIter: integer;
    FMyPrec: double;

    FAFinished: boolean;
  public
    constructor Create(MyVars: TRoot);
    procedure Execute; override;

    property AFinished: boolean read FAFinished write FAFinished;
    property AResult: double read FMyResult write FMyResult;
    property AIter: integer read FMyIter write FMyIter;
    property APrec: double read FMyPrec write FMyPrec;
  end;

  { TMainForm }

  TMainForm = class(TForm)
    CountWithoutCritSecButton: TButton;
    CountWithCritSecButton: TButton;
    LogLBx: TListBox;
    ResultStrGrd: TStringGrid;

    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);

    procedure CountWithoutCritSecButtonClick(Sender: TObject);
    procedure CountWithCritSecButtonClick(Sender: TObject);

  private
  public
    CriticalSection: TCriticalSection;
    Counter: integer;
    UseCriticalSection: boolean;
    procedure DoCounting;
  end;



var
  MainForm: TMainForm;

//  Result: array[1..4] of TResult;

//  MyVars: TRoot;


implementation

{$R *.lfm}

uses
  DateUtils;

const
  NoRoots = 25;
  MaxCalcThreads = 3;


{ TMainForm }
procedure TMainForm.FormCreate(Sender: TObject);
begin
  Logger := TLogger.Create;
  Logger.LogLBx := LogLBx;
  Logger.LogLevel := llDebug;
  Logger.LogMsg(llFatal,'--- TEST THE LOGGER ---');


  ResultStrGrd.Clean;
  ResultStrGrd.FixedCols := 1;
  ResultStrGrd.FixedRows := 1;
  ResultStrGrd.ColCount := 5;
  ResultStrGrd.RowCount := NoRoots + 1;

  // total: 500 px
  ResultStrGrd.ColWidths[0] := 30;
  ResultStrGrd.ColWidths[1] := 120;
  ResultStrGrd.ColWidths[2] := 120;
  ResultStrGrd.ColWidths[3] := 120;
  ResultStrGrd.ColWidths[4] := 50;

  ResultStrGrd.Cells[0,0] := 'No';
  ResultStrGrd.Cells[1,0] := 'Value';
  ResultStrGrd.Cells[2,0] := 'Root';
  ResultStrGrd.Cells[3,0] := 'Root x Root';
  ResultStrGrd.Cells[4,0] := 'Iter';

end;

procedure TMainForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  Logger.Free;
end;


procedure TMainForm.CountWithCritSecButtonClick(Sender: TObject);
begin
  UseCriticalSection := True;
  Logger.LogMsg(llInfo,'WITH Critical Sections');
  DoCounting;
end;


procedure TMainForm.CountWithoutCritSecButtonClick(Sender: TObject);
begin
  UseCriticalSection := False;
  Logger.LogMsg(llInfo,'WITHOUT Critical Sections');
  DoCounting;
end;


procedure TMainForm.DoCounting;
var
  i: integer;
  nlower, nupper, loop: integer;
  LogStr: string;
  Threads: array[1..NoRoots] of TMyThread;
  AllFinished: boolean;
  MyVars: array [1..NoRoots] of TRoot;
  StartTime, FinishTime: TDateTime;
begin
  StartTime := Now;

  // create CriticalSection
  InitializeCriticalSection(CriticalSection);

  // start all threads
  loop := 1;
  NLower := 1;
  NUpper := NLower + MaxCalcThreads;

  repeat
    LogStr := Format('Loop: %d: --- Threads %d - %d',[Loop,NLower,NUpper]);
    Logger.LogMsg(llInfo,LogStr);

    for i:=NLower to NUpper do
    begin
      MyVars[i] := TRoot.Create;
      MyVars[i].prec := 1e-12;
      MyVars[i].xvalue := i;

      // Create the threads
      Threads[i] := TMyThread.Create(MyVars[i]);
      LogStr := Format('Created Thread ID: 0x%x',[Threads[i].ThreadID]);
      Logger.LogMsg(LogStr);
    end;

    // wait until threads are finished
    repeat
      AllFinished := True;

      //for i:=Low(Threads) to High(Threads) do
      for i:=NLower to NUpper do
      begin
        if not Threads[i].AFinished then
          AllFinished := False;

        // if a thread already finished save the result (repeated again and again)
        if Threads[i].AFinished then
        begin
          ResultStrGrd.Cells[0,i] := IntToStr(i);
          ResultStrGrd.Cells[1,i] := FloatToStr(Threads[i].FMyNumber);
          ResultStrGrd.Cells[2,i] := FloatToStr(Threads[i].AResult);
          ResultStrGrd.Cells[3,i] := FloatToStr(Threads[i].AResult * Threads[i].AResult);
          ResultStrGrd.Cells[4,i] := IntToStr(Threads[i].AIter);
        end;
      end;

    until AllFinished;

    loop := loop +1;
    NLower := NLower + MaxCalcThreads + 1;
    NUpper := NLower + MaxCalcThreads;
    if NUpper > NoRoots then
      NUpper := NoRoots;

  until NLower > NoRoots;

  // free all threads
  for i:=1 to NoRoots do
    Threads[i].Free;

  // free the critical section
  DeleteCriticalSection(CriticalSection);

  FinishTime := Now;
  LogStr := Format('Finished: time used: %d sec',[SecondsBetween(StartTime,FinishTime)]);
  Logger.LogMsg(LogSTr);

  // don't forget to Free these
  for i:=1 to NoRoots do
    MyVars[i].Free;

end;


{ MyThread }
constructor TMyThread.Create(MyVars: TRoot);
begin
  inherited Create(False);   // CreateSuspended = False;

//  FreeOnTerminate := True;

  // put job parameter in the thread variables
  FMyNumber := MyVars.xvalue;
  FMyPrec := MyVars.prec;

end;


procedure TMyThread.Execute;
var
  MyVars: TRoot;
  LogSTr: string;
begin
  FAFinished := False;

  if MainForm.UseCriticalSection then
    EnterCriticalSection(MainForm.CriticalSection);

  MyVars := TRoot.Create;
  try
    MyVars.xvalue := FMyNumber;
    MyVars.prec := FMyPrec;

    IterateRoot(MyVars);

    // save the results in the thread structure
    FMyResult := MyVars.root;
    FMyIter := MyVars.iterations;

    LogStr := Format('Hallo! Calling from Thread 0x%x',[self.ThreadID]);
    Logger.LogMsg(LogSTr);

  finally
    MyVars.Free;
    if MainForm.UseCriticalSection then
      LeaveCriticalSection(MainForm.CriticalSection);
  end;

  FAFinished := True;
end;

end.

